#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.cloud.hooks.ray import RayJobHook

from unit.google.cloud.utils.base_gcp_mock import (
    mock_base_gcp_hook_default_project_id,
)

TEST_GCP_CONN_ID: str = "test-gcp-conn-id"
TEST_CLUSTER_NAME: str = "test-cluster-name"

BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
RAY_JOB_STRING = "airflow.providers.google.cloud.hooks.ray.{}"

TEST_CLUSTER_HOSTNAME = "ray.aiplatform-training.googleusercontent.com"
TEST_JOB_ID = "test-job-id"


class TestRayJobHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = RayJobHook(gcp_conn_id=TEST_GCP_CONN_ID)
            self.hook.get_credentials = mock.MagicMock()

    @mock.patch(RAY_JOB_STRING.format("JobSubmissionClient"))
    def test_submit_job(self, mock_client_cls) -> None:
        mock_client = mock_client_cls.return_value
        mock_client.submit_job.return_value = TEST_JOB_ID
        job_id = self.hook.submit_job(
            entrypoint="python3 heavy.py",
            cluster_address=TEST_CLUSTER_HOSTNAME,
            runtime_env={"a": "b"},
            metadata={"k": "v"},
            submission_id="sub-123",
            entrypoint_num_cpus=1,
            entrypoint_num_gpus=0,
            entrypoint_memory=1024,
            entrypoint_resources={"CPU": 1.0},
        )

        mock_client_cls.assert_called_once_with(f"vertex_ray://{TEST_CLUSTER_HOSTNAME}")
        mock_client.submit_job.assert_called_once_with(
            entrypoint="python3 heavy.py",
            runtime_env={"a": "b"},
            metadata={"k": "v"},
            submission_id="sub-123",
            entrypoint_num_cpus=1,
            entrypoint_num_gpus=0,
            entrypoint_memory=1024,
            entrypoint_resources={"CPU": 1.0},
        )
        assert job_id == TEST_JOB_ID

    @mock.patch(RAY_JOB_STRING.format("JobSubmissionClient"))
    def test_stop_job(self, mock_client_cls) -> None:
        mock_client = mock_client_cls.return_value
        mock_client.stop_job.return_value = True

        result = self.hook.stop_job(
            job_id=TEST_JOB_ID,
            cluster_address=TEST_CLUSTER_HOSTNAME,
        )

        mock_client_cls.assert_called_once_with(f"vertex_ray://{TEST_CLUSTER_HOSTNAME}")
        mock_client.stop_job.assert_called_once_with(job_id=TEST_JOB_ID)
        assert result is True

    @mock.patch(RAY_JOB_STRING.format("JobSubmissionClient"))
    def test_delete_job(self, mock_client_cls) -> None:
        mock_client = mock_client_cls.return_value
        mock_client.delete_job.return_value = True

        result = self.hook.delete_job(
            job_id=TEST_JOB_ID,
            cluster_address=TEST_CLUSTER_HOSTNAME,
        )

        mock_client_cls.assert_called_once_with(f"vertex_ray://{TEST_CLUSTER_HOSTNAME}")
        mock_client.delete_job.assert_called_once_with(job_id=TEST_JOB_ID)
        assert result is True

    @mock.patch(RAY_JOB_STRING.format("JobSubmissionClient"))
    def test_get_job_info(self, mock_client_cls) -> None:
        mock_client = mock_client_cls.return_value
        fake_job_details = object()
        mock_client.get_job_info.return_value = fake_job_details

        result = self.hook.get_job_info(
            job_id=TEST_JOB_ID,
            cluster_address=TEST_CLUSTER_HOSTNAME,
        )

        mock_client_cls.assert_called_once_with(f"vertex_ray://{TEST_CLUSTER_HOSTNAME}")
        mock_client.get_job_info.assert_called_once_with(job_id=TEST_JOB_ID)
        assert result is fake_job_details

    @mock.patch(RAY_JOB_STRING.format("JobSubmissionClient"))
    def test_list_jobs(self, mock_client_cls) -> None:
        mock_client = mock_client_cls.return_value
        fake_list = [object(), object()]
        mock_client.list_jobs.return_value = fake_list

        result = self.hook.list_jobs(
            cluster_address=TEST_CLUSTER_HOSTNAME,
        )

        mock_client_cls.assert_called_once_with(f"vertex_ray://{TEST_CLUSTER_HOSTNAME}")
        mock_client.list_jobs.assert_called_once_with()
        assert result is fake_list

    @mock.patch(RAY_JOB_STRING.format("JobSubmissionClient"))
    def test_get_job_status(self, mock_client_cls) -> None:
        mock_client = mock_client_cls.return_value
        fake_status = object()
        mock_client.get_job_status.return_value = fake_status

        result = self.hook.get_job_status(
            job_id=TEST_JOB_ID,
            cluster_address=TEST_CLUSTER_HOSTNAME,
        )

        mock_client_cls.assert_called_once_with(f"vertex_ray://{TEST_CLUSTER_HOSTNAME}")
        mock_client.get_job_status.assert_called_once_with(job_id=TEST_JOB_ID)
        assert result is fake_status

    @mock.patch(RAY_JOB_STRING.format("JobSubmissionClient"))
    def test_get_job_logs(self, mock_client_cls) -> None:
        mock_client = mock_client_cls.return_value
        fake_logs = "some logs"
        mock_client.get_job_logs.return_value = fake_logs

        result = self.hook.get_job_logs(
            job_id=TEST_JOB_ID,
            cluster_address=TEST_CLUSTER_HOSTNAME,
        )

        mock_client_cls.assert_called_once_with(f"vertex_ray://{TEST_CLUSTER_HOSTNAME}")
        mock_client.get_job_logs.assert_called_once_with(job_id=TEST_JOB_ID)
        assert result == fake_logs



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Ray Job operators."""

from __future__ import annotations

import time
from collections.abc import Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any

from google.api_core.exceptions import NotFound
from ray.dashboard.modules.job.common import JobStatus

from airflow.providers.common.compat.sdk import AirflowNotFoundException, AirflowTaskTimeout
from airflow.providers.google.cloud.hooks.ray import RayJobHook
from airflow.providers.google.cloud.links.ray import RayJobLink
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


TERMINAL_STATUSES = {JobStatus.SUCCEEDED.value, JobStatus.FAILED.value}


class OperationFailedException(Exception):
    """Custom exception to handle failing operations on Jobs."""

    pass


class RayJobBaseOperator(GoogleCloudBaseOperator):
    """
    Base class for Jobs on Ray operators.

    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "gcp_conn_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    @cached_property
    def hook(self) -> RayJobHook:
        return RayJobHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )


class RaySubmitJobOperator(RayJobBaseOperator):
    """
    Submit and execute Job on Ray cluster.

    When a job is submitted, it runs once to completion or failure. Retries or
    different runs with different parameters should be handled by the
    submitter. Jobs are bound to the lifetime of a Ray cluster, so if the
    cluster goes down, all running jobs on that cluster will be terminated.

    :param cluster_address: Required. Either (1) the address of the Ray cluster, or (2) the HTTP address
        of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
        In case (1) it must be specified as an address that can be passed to
        ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
        or "auto", or "localhost:<port>".
    :param entrypoint: Required. The shell command to run for this job.
    :param get_job_logs: If set to True, the operator will wait until the end of
        Job execution and output the logs.
    :param wait_for_job_done: If set to True, the operator will wait until the end of
        Job execution. Please note, that if the Job will fail during execution and
        this parameter is set to False, there will be no indication of the failure.
    :param submission_id: A unique ID for this job.
    :param runtime_env: The runtime environment to install and run this job in.
    :param metadata: Arbitrary data to store along with this job.
    :param entrypoint_num_cpus: The quantity of CPU cores to reserve for the execution
        of the entrypoint command, separately from any tasks or actors launched
        by it. Defaults to 0.
    :param entrypoint_num_gpus: The quantity of GPUs to reserve for the execution
        of the entrypoint command, separately from any tasks or actors launched
        by it. Defaults to 0.
    :param entrypoint_memory: The quantity of memory to reserve for the
        execution of the entrypoint command, separately from any tasks or
        actors launched by it. Defaults to 0.
    :param entrypoint_resources: The quantity of custom resources to reserve for the
        execution of the entrypoint command, separately from any tasks or
        actors launched by it.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"entrypoint", "submission_id", "cluster_address"} | set(RayJobBaseOperator.template_fields)
    )
    operator_extra_links = (RayJobLink(),)

    def __init__(
        self,
        cluster_address: str,
        entrypoint: str,
        get_job_logs: bool | None = False,
        wait_for_job_done: bool | None = False,
        runtime_env: dict[str, Any] | None = None,
        metadata: dict[str, str] | None = None,
        submission_id: str | None = None,
        entrypoint_num_cpus: int | float | None = None,
        entrypoint_num_gpus: int | float | None = None,
        entrypoint_memory: int | None = None,
        entrypoint_resources: dict[str, float] | None = None,
        submit_job_timeout: float = 60 * 30,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.cluster_address = cluster_address
        self.get_job_logs = get_job_logs
        self.wait_for_job_done = wait_for_job_done
        self.entrypoint = entrypoint
        self.runtime_env = runtime_env
        self.metadata = metadata
        self.submission_id = submission_id
        self.entrypoint_num_cpus = entrypoint_num_cpus
        self.entrypoint_num_gpus = entrypoint_num_gpus
        self.entrypoint_memory = entrypoint_memory
        self.entrypoint_resources = entrypoint_resources
        self.submit_job_timeout = submit_job_timeout

    def _check_job_status(
        self, cluster_address: str, job_id: str, timeout: float, polling_interval: float = 5.0
    ) -> str:
        "Check if the Job has reached terminated state."
        start = time.monotonic()

        while True:
            job_status = self.hook.get_job_status(cluster_address=cluster_address, job_id=job_id)
            if job_status in TERMINAL_STATUSES:
                self.log.info("Job has finished execution with status: %s", job_status)
                return job_status

            if time.monotonic() - start > timeout:
                raise AirflowTaskTimeout(
                    f"Timeout waiting for Ray Job {job_id} to finish. Last status: {job_status}"
                )

            self.log.info("Job status: %s...", job_status)
            time.sleep(polling_interval)

    def _get_job_logs(self, cluster_address, job_id):
        "Output Job logs."
        logs = self.hook.get_job_logs(cluster_address=cluster_address, job_id=job_id)
        self.log.info("Got job logs:\n%s\n", logs)

    def execute(self, context: Context):
        if self.get_job_logs and not self.wait_for_job_done:
            raise ValueError(
                "Retrieving Job logs can be possible only after Job completion. "
                "Please, enable wait_for_job_done parameter to be able to get logs."
            )
        try:
            self.log.info("Submitting Job on a Ray cluster...")
            submitted_job_id = self.hook.submit_job(
                cluster_address=self.cluster_address,
                entrypoint=self.entrypoint,
                runtime_env=self.runtime_env,
                metadata=self.metadata,
                submission_id=self.submission_id,
                entrypoint_num_cpus=self.entrypoint_num_cpus,
                entrypoint_num_gpus=self.entrypoint_num_gpus,
                entrypoint_memory=self.entrypoint_memory,
                entrypoint_resources=self.entrypoint_resources,
            )
            self.log.info("Submitted Ray Job id=%s", submitted_job_id)
            RayJobLink.persist(
                context=context,
                cluster_address=self.cluster_address,
                job_id=submitted_job_id,
            )
        except RuntimeError as exc:
            raise exc
        if self.wait_for_job_done:
            self._check_job_status(
                cluster_address=self.cluster_address, job_id=submitted_job_id, timeout=self.submit_job_timeout
            )

        if self.get_job_logs:
            self._get_job_logs(cluster_address=self.cluster_address, job_id=submitted_job_id)

        return submitted_job_id


class RayStopJobOperator(RayJobBaseOperator):
    """
    Stop Job on Ray cluster.

    :param cluster_address: Required. Either (1) the address of the Ray cluster, or (2) the HTTP address
        of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
        In case (1) it must be specified as an address that can be passed to
        ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
        or "auto", or "localhost:<port>".
    :param job_id: Required. The job ID or submission ID for the job to be stopped.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"cluster_address", "job_id"} | set(RayJobBaseOperator.template_fields)
    )
    operator_extra_links = ()

    def __init__(
        self,
        cluster_address: str,
        job_id: str,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.cluster_address = cluster_address
        self.job_id = job_id

    def execute(self, context: Context):
        self.log.info("Stopping Job %s on a Ray cluster...", self.job_id)

        try:
            is_stopped = self.hook.stop_job(
                cluster_address=self.cluster_address,
                job_id=self.job_id,
            )
        except NotFound:
            raise AirflowNotFoundException("Job with specified id was not found on cluster.")

        if is_stopped:
            self.log.info("Job was successfully stopped.")
            return

        raise OperationFailedException("Some error happened during stopping the Job. Exiting.")


class RayDeleteJobOperator(RayJobBaseOperator):
    """
    Delete Job on Ray cluster in a terminal state and all of its associated data.

    If the job is not already in a terminal state, raises an error.
    This does not delete the job logs from disk.
    Submitting a job with the same submission ID as a previously
    deleted job is not supported and may lead to unexpected behavior.

    :param cluster_address: Required. Either (1) the address of the Ray cluster, or (2) the HTTP address
        of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
        In case (1) it must be specified as an address that can be passed to
        ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
        or "auto", or "localhost:<port>".
    :param job_id: Required. The job ID or submission ID for the job to be stopped.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"cluster_address", "job_id"} | set(RayJobBaseOperator.template_fields)
    )
    operator_extra_links = ()

    def __init__(
        self,
        cluster_address: str,
        job_id: str,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.cluster_address = cluster_address
        self.job_id = job_id

    def execute(self, context: Context):
        self.log.info("Deleting Job %s on a Ray cluster...", self.job_id)

        try:
            is_deleted = self.hook.delete_job(
                cluster_address=self.cluster_address,
                job_id=self.job_id,
            )
        except NotFound:
            raise AirflowNotFoundException("Job with specified id was not found on cluster.")

        if is_deleted:
            self.log.info("Job was successfully deleted.")
            return

        raise OperationFailedException("Some error happened during deleting the Job. Exiting.")


class RayGetJobInfoOperator(RayJobBaseOperator):
    """
    Get the latest status and other information associated with a Job on Ray cluster.

    :param cluster_address: Required. Either (1) the address of the Ray cluster, or (2) the HTTP address
        of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
        In case (1) it must be specified as an address that can be passed to
        ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
        or "auto", or "localhost:<port>".
    :param job_id: Required. The job ID or submission ID for the job to be stopped.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple(
        {"cluster_address", "job_id"} | set(RayJobBaseOperator.template_fields)
    )
    operator_extra_links = ()

    def __init__(
        self,
        cluster_address: str,
        job_id: str,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.cluster_address = cluster_address
        self.job_id = job_id

    def execute(self, context: Context):
        self.log.info("Retrieving information about Job %s on a Ray cluster...", self.job_id)

        try:
            job_info = self.hook.get_job_info(
                cluster_address=self.cluster_address,
                job_id=self.job_id,
            )
        except NotFound:
            raise AirflowNotFoundException("Job with specified id was not found on cluster.")

        # Everything below is outside the try/except so serialization/logging errors are not masked
        self.log.info("Job information:\n %s \n", job_info)

        ray_job_dict = self.hook.serialize_job_obj(job_info)
        return ray_job_dict


class RayListJobsOperator(RayJobBaseOperator):
    """
    List all jobs along with their status and other information.

    Lists all jobs that have ever run on the cluster, including jobs that are
    currently running and jobs that are no longer running.

    :param cluster_address: Required. Either (1) the address of the Ray cluster, or (2) the HTTP address
        of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
        In case (1) it must be specified as an address that can be passed to
        ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
        or "auto", or "localhost:<port>".
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = tuple({"cluster_address"} | set(RayJobBaseOperator.template_fields))
    operator_extra_links = ()

    def __init__(
        self,
        cluster_address: str,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.cluster_address = cluster_address

    def execute(self, context: Context):
        self.log.info("Listing Jobs on a Ray cluster...")
        jobs = self.hook.list_jobs(
            cluster_address=self.cluster_address,
        )
        if jobs:
            self.log.info("Outputting first 10 Jobs...")
            for job in jobs[:10]:
                self.log.info("Job:\n %s \n \n", job)
            list_jobs = [self.hook.serialize_job_obj(job) for job in jobs]
            return list_jobs
        self.log.info("No Jobs found.")
        return



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that uses Google AutoML services.
"""

from __future__ import annotations

import os
from datetime import datetime

from google.cloud.aiplatform import schema
from google.protobuf.struct_pb2 import Value

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.vertex_ai.auto_ml import (
    CreateAutoMLImageTrainingJobOperator,
    DeleteAutoMLTrainingJobOperator,
)
from airflow.providers.google.cloud.operators.vertex_ai.dataset import (
    CreateDatasetOperator,
    DeleteDatasetOperator,
    ImportDataOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
DAG_ID = "automl_vision_obj_detect"
REGION = "us-central1"
IMAGE_DISPLAY_NAME = f"automl-vision-detect-{ENV_ID}"
MODEL_DISPLAY_NAME = f"automl-vision-detect-model-{ENV_ID}"
RESOURCE_DATA_BUCKET = "airflow-system-tests-resources"

IMAGE_DATASET = {
    "display_name": f"image-detect-dataset-{ENV_ID}",
    "metadata_schema_uri": schema.dataset.metadata.image,
    "metadata": Value(string_value="image-dataset"),
}

IMAGE_DATA_CONFIG = [
    # For testing only
    {
        "import_schema_uri": schema.dataset.ioformat.image.bounding_box,
        "gcs_source": {
            "uris": [f"gs://{RESOURCE_DATA_BUCKET}/automl/datasets/vision/obj_detection_short.csv"]
        },
    },
]

# Example DAG for AutoML Vision Object Detection
with DAG(
    DAG_ID,
    schedule="@once",  # Override to match your needs
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "vertex_ai", "automl", "vision", "object-detection"],
) as dag:
    create_image_dataset = CreateDatasetOperator(
        task_id="image_dataset",
        dataset=IMAGE_DATASET,
        region=REGION,
        project_id=PROJECT_ID,
    )
    image_dataset_id = create_image_dataset.output["dataset_id"]

    import_image_dataset = ImportDataOperator(
        task_id="import_image_data",
        dataset_id=image_dataset_id,
        region=REGION,
        project_id=PROJECT_ID,
        import_configs=IMAGE_DATA_CONFIG,
    )
    # [START how_to_cloud_vertex_ai_create_auto_ml_image_object_detection_training_job_operator]
    create_auto_ml_image_training_job = CreateAutoMLImageTrainingJobOperator(
        task_id="auto_ml_image_task",
        display_name=IMAGE_DISPLAY_NAME,
        dataset_id=image_dataset_id,
        prediction_type="object_detection",
        multi_label=False,
        model_type="CLOUD",
        training_fraction_split=0.6,
        validation_fraction_split=0.2,
        test_fraction_split=0.2,
        budget_milli_node_hours=20000,
        model_display_name=MODEL_DISPLAY_NAME,
        disable_early_stopping=False,
        region=REGION,
        project_id=PROJECT_ID,
    )
    # [END how_to_cloud_vertex_ai_create_auto_ml_image_object_detection_training_job_operator]

    delete_auto_ml_image_training_job = DeleteAutoMLTrainingJobOperator(
        task_id="delete_auto_ml_training_job",
        training_pipeline_id="{{ task_instance.xcom_pull(task_ids='auto_ml_image_task', "
        "key='training_id') }}",
        region=REGION,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_image_dataset = DeleteDatasetOperator(
        task_id="delete_image_dataset",
        dataset_id=image_dataset_id,
        region=REGION,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        # TEST SETUP
        create_image_dataset
        >> import_image_dataset
        # TEST BODY
        >> create_auto_ml_image_training_job
        # TEST TEARDOWN
        >> delete_auto_ml_image_training_job
        >> delete_image_dataset
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Ray Job hook."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any
from urllib.parse import urlparse

from ray.job_submission import JobSubmissionClient

from airflow.providers.google.common.hooks.base_google import GoogleBaseHook

if TYPE_CHECKING:
    from ray.dashboard.modules.job.common import JobStatus
    from ray.dashboard.modules.job.pydantic_models import JobDetails

VERTEX_RAY_DOMAIN = "aiplatform-training.googleusercontent.com"


class RayJobHook(GoogleBaseHook):
    """Hook for Jobs APIs."""

    def _is_vertex_ray_address(self, address: str) -> bool:
        """Return True if address points to Vertex Ray dashboard host."""
        parsed = urlparse(address if "://" in address else f"https://{address}")
        hostname = parsed.hostname
        if not hostname:
            return False
        return hostname.endswith(VERTEX_RAY_DOMAIN)

    def get_client(self, address: str) -> JobSubmissionClient:
        """
        Create a client for submitting and interacting with jobs on a remote cluster.

        :param address: Either (1) the address of the Ray cluster, or (2) the HTTP address
            of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
            In case (1) it must be specified as an address that can be passed to
            ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
            or "auto", or "localhost:<port>".
        """
        if self._is_vertex_ray_address(address):
            return JobSubmissionClient(f"vertex_ray://{address}")
        return JobSubmissionClient(address=address)

    def serialize_job_obj(self, job_obj: JobDetails) -> dict:
        """Serialize JobDetails to a plain dict."""
        if hasattr(job_obj, "model_dump"):  # Pydantic v2
            return job_obj.model_dump(exclude_none=True)
        if hasattr(job_obj, "dict"):  # Pydantic v1
            return job_obj.dict(exclude_none=True)
        return dict(job_obj)

    def submit_job(
        self,
        entrypoint: str,
        cluster_address: str,
        runtime_env: dict[str, Any] | None = None,
        metadata: dict[str, str] | None = None,
        submission_id: str | None = None,
        entrypoint_num_cpus: int | float | None = None,
        entrypoint_num_gpus: int | float | None = None,
        entrypoint_memory: int | None = None,
        entrypoint_resources: dict[str, float] | None = None,
    ) -> str:
        """
        Submit and execute Job on Ray cluster.

        When a job is submitted, it runs once to completion or failure. Retries or
        different runs with different parameters should be handled by the
        submitter. Jobs are bound to the lifetime of a Ray cluster, so if the
        cluster goes down, all running jobs on that cluster will be terminated.

        :param entrypoint: Required. The shell command to run for this job.
        :param cluster_address: Required. Either (1) the address of the Ray cluster, or (2) the HTTP address
            of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
            In case (1) it must be specified as an address that can be passed to
            ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
            or "auto", or "localhost:<port>".
        :param submission_id: A unique ID for this job.
        :param runtime_env: The runtime environment to install and run this job in.
        :param metadata: Arbitrary data to store along with this job.
        :param entrypoint_num_cpus: The quantity of CPU cores to reserve for the execution
            of the entrypoint command, separately from any tasks or actors launched
            by it. Defaults to 0.
        :param entrypoint_num_gpus: The quantity of GPUs to reserve for the execution
            of the entrypoint command, separately from any tasks or actors launched
            by it. Defaults to 0.
        :param entrypoint_memory: The quantity of memory to reserve for the
            execution of the entrypoint command, separately from any tasks or
            actors launched by it. Defaults to 0.
        :param entrypoint_resources: The quantity of custom resources to reserve for the
            execution of the entrypoint command, separately from any tasks or
            actors launched by it.
        """
        job_id = self.get_client(address=cluster_address).submit_job(
            entrypoint=entrypoint,
            runtime_env=runtime_env,
            metadata=metadata,
            submission_id=submission_id,
            entrypoint_num_cpus=entrypoint_num_cpus,
            entrypoint_num_gpus=entrypoint_num_gpus,
            entrypoint_memory=entrypoint_memory,
            entrypoint_resources=entrypoint_resources,
        )
        return job_id

    def stop_job(
        self,
        job_id: str,
        cluster_address: str,
    ) -> bool:
        """
        Stop Job on Ray cluster.

        :param job_id: Required. The job ID or submission ID for the job to be stopped.
        :param cluster_address: Required. Either (1) the address of the Ray cluster, or (2) the HTTP address
            of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
            In case (1) it must be specified as an address that can be passed to
            ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
            or "auto", or "localhost:<port>".
        :return: True if the job was stopped, otherwise False.
        """
        return self.get_client(address=cluster_address).stop_job(job_id=job_id)

    def delete_job(
        self,
        job_id: str,
        cluster_address: str,
    ) -> bool:
        """
        Delete Job on Ray cluster in a terminal state and all of its associated data.

        If the job is not already in a terminal state, raises an error.
        This does not delete the job logs from disk.
        Submitting a job with the same submission ID as a previously
        deleted job is not supported and may lead to unexpected behavior.

        :param job_id: Required. The job ID or submission ID for the job to be deleted.
        :param cluster_address: Required. Either (1) the address of the Ray cluster, or (2) the HTTP address
            of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
            In case (1) it must be specified as an address that can be passed to
            ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
            or "auto", or "localhost:<port>".
        :return: True if the job was deleted, otherwise False.
        """
        return self.get_client(address=cluster_address).delete_job(job_id=job_id)

    def get_job_info(
        self,
        job_id: str,
        cluster_address: str,
    ) -> JobDetails:
        """
        Get the latest status and other information associated with a Job on Ray cluster.

        :param job_id: Required. The job ID or submission ID for the job to be retrieved.
        :param cluster_address: Required. Either (1) the address of the Ray cluster, or (2) the HTTP address
            of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
            In case (1) it must be specified as an address that can be passed to
            ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
            or "auto", or "localhost:<port>".
        :return: The JobDetails for the job.
        """
        return self.get_client(address=cluster_address).get_job_info(job_id=job_id)

    def list_jobs(
        self,
        cluster_address: str,
    ) -> list[JobDetails]:
        """
        List all jobs along with their status and other information.

        Lists all jobs that have ever run on the cluster, including jobs that are
        currently running and jobs that are no longer running.

        :param cluster_address: Required. Either (1) the address of the Ray cluster, or (2) the HTTP address
            of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
            In case (1) it must be specified as an address that can be passed to
            ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
            or "auto", or "localhost:<port>".
        """
        return self.get_client(address=cluster_address).list_jobs()

    def get_job_status(
        self,
        job_id: str,
        cluster_address: str,
    ) -> JobStatus:
        """
        Get the most recent status of a Job on Ray cluster.

        :param job_id: Required. The job ID or submission ID for the job to be retrieved.
        :param cluster_address: Required. Either (1) the address of the Ray cluster, or (2) the HTTP address
            of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
            In case (1) it must be specified as an address that can be passed to
            ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
            or "auto", or "localhost:<port>".
        :return: The JobStatus of the job.
        """
        return self.get_client(address=cluster_address).get_job_status(job_id=job_id)

    def get_job_logs(
        self,
        job_id: str,
        cluster_address: str,
    ) -> str:
        """
        Get all logs produced by a Job on Ray cluster.

        :param job_id: Required. The job ID or submission ID for the job to be retrieved.
        :param cluster_address: Required. Either (1) the address of the Ray cluster, or (2) the HTTP address
            of the dashboard server on the head node, e.g. "http://<head-node-ip>:8265".
            In case (1) it must be specified as an address that can be passed to
            ray.init(), e.g. a Ray Client address (ray://<head_node_host>:10001),
            or "auto", or "localhost:<port>".
        :return: A string containing the full logs of the job.
        """
        return self.get_client(address=cluster_address).get_job_logs(job_id=job_id)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import asyncio
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.providers.google.cloud.hooks.gen_ai import BatchJobStatus, GenAIGeminiAPIAsyncHook
from airflow.triggers.base import BaseTrigger, TriggerEvent

if TYPE_CHECKING:
    from google.genai.types import CreateBatchJobConfig


class GenAIGeminiCreateBatchJobTrigger(BaseTrigger):
    """
    Trigger that creates Gemini Batch Job and waiting for execution.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param model: Required. The name of the publisher model to use for Batch job.
    :param gemini_api_key: Required. Key to interact with Gemini Batch API.
    :param input_source: Required. Source of requests, could be inline requests or file name.
    :param results_folder: Optional. Path to a folder on local machine where file with results will be saved.
    :param create_batch_job_config: Optional. Config for batch job creation.
    :param retrieve_result: Optional. Push the result to XCom. If the input_source is inline, this pushes
        the execution result. If a file name is specified, this pushes the output file path.
    :param polling_interval: Optional. The interval, in seconds, to poll the job status.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    def __init__(
        self,
        project_id: str,
        location: str,
        model: str,
        input_source: list | str,
        gemini_api_key: str,
        create_batch_job_config: CreateBatchJobConfig | dict | None = None,
        results_folder: str | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        retrieve_result: bool = False,
        polling_interval: int = 30,
    ):
        super().__init__()

        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.model = model
        self.input_source = input_source
        self.create_batch_job_config = create_batch_job_config
        self.gemini_api_key = gemini_api_key
        self.retrieve_result = retrieve_result
        self.polling_interval = polling_interval
        self.results_folder = results_folder

    def _get_async_hook(self) -> GenAIGeminiAPIAsyncHook:
        return GenAIGeminiAPIAsyncHook(gemini_api_key=self.gemini_api_key)

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize class arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.gen_ai.GenAIGeminiCreateBatchJobTrigger",
            {
                "project_id": self.project_id,
                "location": self.location,
                "gcp_conn_id": self.gcp_conn_id,
                "impersonation_chain": self.impersonation_chain,
                "model": self.model,
                "input_source": self.input_source,
                "create_batch_job_config": self.create_batch_job_config,
                "gemini_api_key": self.gemini_api_key,
                "retrieve_result": self.retrieve_result,
                "polling_interval": self.polling_interval,
                "results_folder": self.results_folder,
            },
        )

    async def run(self):
        """
        Loop until the job reaches  successful final or error state.

        Yields a TriggerEvent with success status, if the job reaches successful state.

        Yields a TriggerEvent with error status, if the client returns an unexpected terminal
        job status or any exception is raised while looping.

        In any other case the Trigger will wait for a specified amount of time
        stored in self.polling_interval variable.
        """
        try:
            hook = self._get_async_hook()
            job = await hook.create_batch_job(
                model=self.model,
                source=self.input_source,
                create_batch_job_config=self.create_batch_job_config,
            )
            while True:
                job = await hook.get_batch_job(job_name=job.name)
                if job.state.name == BatchJobStatus.SUCCEEDED.value:
                    self.log.info("Job execution completed")
                    yield TriggerEvent(
                        {
                            "status": "success",
                            "message": "Job completed",
                            "job_name": job.name,
                        }
                    )
                    return
                if job.state.name in [
                    BatchJobStatus.FAILED.value,
                    BatchJobStatus.EXPIRED.value,
                    BatchJobStatus.CANCELLED.value,
                ]:
                    self.log.error("Job execution was not completed!")
                    yield TriggerEvent(
                        {"status": "error", "message": f"Job {job.name} execution was not completed!"}
                    )
                    return
                self.log.info(
                    "Waiting for job execution, polling interval: %s seconds, current state: %s",
                    self.polling_interval,
                    job.state.name,
                )
                await asyncio.sleep(self.polling_interval)

        except Exception as e:
            self.log.exception("Exception occurred while checking for job completion.")
            yield TriggerEvent({"status": "error", "message": str(e)})


class GenAIGeminiCreateEmbeddingsBatchJobTrigger(BaseTrigger):
    """
    Trigger that creates Gemini Embeddings Batch Job and waiting for execution.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param model: Required. The name of the publisher model to use for Batch job.
    :param gemini_api_key: Required. Key to interact with Gemini Batch API.
    :param input_source: Required. Source of requests, could be inline requests or file name.
    :param results_folder: Optional. Path to a folder on local machine where file with results will be saved.
    :param create_embeddings_config: Optional. Config for batch job creation.
    :param retrieve_result: Optional. Push the result to XCom. If the input_source is inline, this pushes
        the execution result. If a file name is specified, this pushes the output file path.
    :param polling_interval: Optional. The interval, in seconds, to poll the job status.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    def __init__(
        self,
        project_id: str,
        location: str,
        model: str,
        input_source: dict | str,
        gemini_api_key: str,
        create_embeddings_config: CreateBatchJobConfig | dict | None = None,
        results_folder: str | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        retrieve_result: bool = False,
        polling_interval: int = 30,
    ):
        super().__init__()

        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.model = model
        self.input_source = input_source
        self.create_embeddings_config = create_embeddings_config
        self.gemini_api_key = gemini_api_key
        self.retrieve_result = retrieve_result
        self.polling_interval = polling_interval
        self.results_folder = results_folder

    def _get_async_hook(self) -> GenAIGeminiAPIAsyncHook:
        return GenAIGeminiAPIAsyncHook(gemini_api_key=self.gemini_api_key)

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize class arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.gen_ai.GenAIGeminiCreateEmbeddingsBatchJobTrigger",
            {
                "project_id": self.project_id,
                "location": self.location,
                "gcp_conn_id": self.gcp_conn_id,
                "impersonation_chain": self.impersonation_chain,
                "model": self.model,
                "input_source": self.input_source,
                "create_embeddings_config": self.create_embeddings_config,
                "gemini_api_key": self.gemini_api_key,
                "retrieve_result": self.retrieve_result,
                "polling_interval": self.polling_interval,
                "results_folder": self.results_folder,
            },
        )

    async def run(self):
        """
        Loop until the job reaches  successful final or error state.

        Yields a TriggerEvent with success status, if the job reaches successful state.

        Yields a TriggerEvent with error status, if the client returns an unexpected terminal
        job status or any exception is raised while looping.

        In any other case the Trigger will wait for a specified amount of time
        stored in self.polling_interval variable.
        """
        try:
            hook = self._get_async_hook()
            job = await hook.create_embeddings_batch_job(
                model=self.model,
                source=self.input_source,
                create_embeddings_config=self.create_embeddings_config,
            )
            while True:
                job = await hook.get_batch_job(job_name=job.name)
                if job.state.name == BatchJobStatus.SUCCEEDED.value:
                    self.log.info("Job execution completed")
                    yield TriggerEvent(
                        {
                            "status": "success",
                            "message": "Job completed",
                            "job": job.model_dump(),
                        }
                    )
                    return
                if job.state.name in [
                    BatchJobStatus.FAILED.value,
                    BatchJobStatus.EXPIRED.value,
                    BatchJobStatus.CANCELLED.value,
                ]:
                    self.log.error("Job execution was not completed!")
                    yield TriggerEvent(
                        {"status": "error", "message": f"Job {job.name} execution was not completed!"}
                    )
                    return
                self.log.info(
                    "Waiting for job execution, polling interval: %s seconds, current state: %s",
                    self.polling_interval,
                    job.state.name,
                )
                await asyncio.sleep(self.polling_interval)

        except Exception as e:
            self.log.exception("Exception occurred while checking for job completion.")
            yield TriggerEvent({"status": "error", "message": str(e)})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.cloud.links.ray import RayJobLink

TEST_CLUSTER_ADDRESS = "ray-head-123.us-central1.ray.googleusercontent.com"
TEST_JOB_ID = "test-job-id"

EXPECTED_RAY_JOB_LINK_NAME = "Ray Job"
EXPECTED_RAY_JOB_LINK_KEY = "ray_job"
EXPECTED_RAY_JOB_LINK_FORMAT_STR = "http://{cluster_address}/#/jobs/{job_id}"


class TestRayJobLink:
    def test_class_attributes(self):
        assert RayJobLink.key == EXPECTED_RAY_JOB_LINK_KEY
        assert RayJobLink.name == EXPECTED_RAY_JOB_LINK_NAME
        assert RayJobLink.format_str == EXPECTED_RAY_JOB_LINK_FORMAT_STR

    def test_persist(self):
        mock_context = mock.MagicMock()
        mock_context["ti"] = mock.MagicMock()
        mock_context["task"] = mock.MagicMock()

        RayJobLink.persist(
            context=mock_context,
            cluster_address=TEST_CLUSTER_ADDRESS,
            job_id=TEST_JOB_ID,
        )

        mock_context["ti"].xcom_push.assert_called_once_with(
            key=EXPECTED_RAY_JOB_LINK_KEY,
            value={
                "cluster_address": TEST_CLUSTER_ADDRESS,
                "job_id": TEST_JOB_ID,
            },
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import PropertyMock

import google
import pytest
from google.cloud.bigtable import Client, enums
from google.cloud.bigtable.instance import Instance

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.bigtable import BigtableHook
from airflow.providers.google.common.consts import CLIENT_INFO

from unit.google.cloud.utils.base_gcp_mock import (
    GCP_PROJECT_ID_HOOK_UNIT_TEST,
    mock_base_gcp_hook_default_project_id,
    mock_base_gcp_hook_no_default_project_id,
)

CBT_INSTANCE = "instance"
CBT_INSTANCE_DISPLAY_NAME = "test instance"
CBT_INSTANCE_TYPE = enums.Instance.Type.PRODUCTION
CBT_INSTANCE_LABELS = {"env": "sit"}
CBT_CLUSTER = "cluster"
CBT_ZONE = "zone"
CBT_TABLE = "table"
CBT_REPLICA_CLUSTER_ID = "replica-cluster"
CBT_REPLICA_CLUSTER_ZONE = "us-west1-b"
CBT_REPLICATE_CLUSTERS = [
    {"id": "replica-1", "zone": "us-west1-a"},
    {"id": "replica-2", "zone": "us-central1-f"},
    {"id": "replica-3", "zone": "us-east1-d"},
]


class TestBigtableHookNoDefaultProjectId:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_no_default_project_id,
        ):
            self.bigtable_hook_no_default_project_id = BigtableHook(gcp_conn_id="test")

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook.get_credentials")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.Client")
    def test_bigtable_client_creation(self, mock_client, mock_get_creds):
        result = self.bigtable_hook_no_default_project_id._get_client(GCP_PROJECT_ID_HOOK_UNIT_TEST)
        mock_client.assert_called_once_with(
            project=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            credentials=mock_get_creds.return_value,
            client_info=CLIENT_INFO,
            admin=True,
        )
        assert mock_client.return_value == result
        assert self.bigtable_hook_no_default_project_id._client == result

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_get_instance_overridden_project_id(self, get_client):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        res = self.bigtable_hook_no_default_project_id.get_instance(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST, instance_id=CBT_INSTANCE
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="example-project")
        assert res is not None

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_instance_overridden_project_id(self, get_client):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        delete_method = instance_method.return_value.delete
        res = self.bigtable_hook_no_default_project_id.delete_instance(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST, instance_id=CBT_INSTANCE
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        delete_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="example-project")
        assert res is None

    @mock.patch("google.cloud.bigtable.instance.Instance.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_instance_overridden_project_id(self, get_client, instance_create):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_create.return_value = operation
        res = self.bigtable_hook_no_default_project_id.create_instance(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            instance_id=CBT_INSTANCE,
            main_cluster_id=CBT_CLUSTER,
            main_cluster_zone=CBT_ZONE,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_create.assert_called_once_with(clusters=mock.ANY)
        assert res.instance_id == "instance"

    @mock.patch("google.cloud.bigtable.instance.Instance.update")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_update_instance_overridden_project_id(self, get_client, instance_update):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_update.return_value = operation
        res = self.bigtable_hook_no_default_project_id.update_instance(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            instance_id=CBT_INSTANCE,
            instance_display_name=CBT_INSTANCE_DISPLAY_NAME,
            instance_type=CBT_INSTANCE_TYPE,
            instance_labels=CBT_INSTANCE_LABELS,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_update.assert_called_once_with()
        assert res.instance_id == "instance"

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_table_overridden_project_id(self, get_client):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        table_delete_method = instance_method.return_value.table.return_value.delete
        instance_exists_method.return_value = True
        self.bigtable_hook_no_default_project_id.delete_table(
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST, instance_id=CBT_INSTANCE, table_id=CBT_TABLE
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_exists_method.assert_called_once_with()
        table_delete_method.assert_called_once_with()


class TestBigtableHookDefaultProjectId:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.bigtable_hook_default_project_id = BigtableHook(gcp_conn_id="test")

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook.get_credentials")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.Client")
    def test_bigtable_client_creation(self, mock_client, mock_get_creds):
        result = self.bigtable_hook_default_project_id._get_client(GCP_PROJECT_ID_HOOK_UNIT_TEST)
        mock_client.assert_called_once_with(
            project=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            credentials=mock_get_creds.return_value,
            client_info=CLIENT_INFO,
            admin=True,
        )
        assert mock_client.return_value == result
        assert self.bigtable_hook_default_project_id._client == result

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_get_instance(self, get_client, mock_project_id):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        res = self.bigtable_hook_default_project_id.get_instance(
            instance_id=CBT_INSTANCE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="example-project")
        assert res is not None

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_get_instance_overridden_project_id(self, get_client):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        res = self.bigtable_hook_default_project_id.get_instance(
            project_id="new-project", instance_id=CBT_INSTANCE
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="new-project")
        assert res is not None

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_get_instance_no_instance(self, get_client, mock_project_id):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = False
        res = self.bigtable_hook_default_project_id.get_instance(
            instance_id=CBT_INSTANCE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="example-project")
        assert res is None

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_instance(self, get_client, mock_project_id):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        delete_method = instance_method.return_value.delete
        res = self.bigtable_hook_default_project_id.delete_instance(
            instance_id=CBT_INSTANCE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        delete_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="example-project")
        assert res is None

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_instance_overridden_project_id(self, get_client):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        delete_method = instance_method.return_value.delete
        res = self.bigtable_hook_default_project_id.delete_instance(
            project_id="new-project", instance_id=CBT_INSTANCE
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        delete_method.assert_called_once_with()
        get_client.assert_called_once_with(project_id="new-project")
        assert res is None

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_instance_no_instance(self, get_client, mock_project_id):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = False
        delete_method = instance_method.return_value.delete
        self.bigtable_hook_default_project_id.delete_instance(
            instance_id=CBT_INSTANCE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        instance_method.assert_called_once_with("instance")
        instance_exists_method.assert_called_once_with()
        delete_method.assert_not_called()
        get_client.assert_called_once_with(project_id="example-project")

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("google.cloud.bigtable.instance.Instance.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_instance(self, get_client, instance_create, mock_project_id):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_create.return_value = operation
        res = self.bigtable_hook_default_project_id.create_instance(
            instance_id=CBT_INSTANCE,
            main_cluster_id=CBT_CLUSTER,
            main_cluster_zone=CBT_ZONE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_create.assert_called_once_with(clusters=mock.ANY)
        assert res.instance_id == "instance"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("google.cloud.bigtable.instance.Instance.cluster")
    @mock.patch("google.cloud.bigtable.instance.Instance.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_instance_with_one_replica_cluster_production(
        self, get_client, instance_create, cluster, mock_project_id
    ):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_create.return_value = operation

        res = self.bigtable_hook_default_project_id.create_instance(
            instance_id=CBT_INSTANCE,
            main_cluster_id=CBT_CLUSTER,
            main_cluster_zone=CBT_ZONE,
            replica_clusters=[{"id": CBT_REPLICA_CLUSTER_ID, "zone": CBT_REPLICA_CLUSTER_ZONE}],
            cluster_nodes=1,
            cluster_storage_type=enums.StorageType.SSD,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            instance_type=enums.Instance.Type.PRODUCTION,
        )
        cluster.assert_has_calls(
            [
                mock.call(
                    cluster_id=CBT_CLUSTER,
                    location_id=CBT_ZONE,
                    serve_nodes=1,
                    default_storage_type=enums.StorageType.SSD,
                ),
                mock.call(CBT_REPLICA_CLUSTER_ID, CBT_REPLICA_CLUSTER_ZONE, 1, enums.StorageType.SSD),
            ],
            any_order=True,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_create.assert_called_once_with(clusters=mock.ANY)
        assert res.instance_id == "instance"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("google.cloud.bigtable.instance.Instance.cluster")
    @mock.patch("google.cloud.bigtable.instance.Instance.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_instance_with_one_replica_cluster_development(
        self, get_client, instance_create, cluster, mock_project_id
    ):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_create.return_value = operation

        res = self.bigtable_hook_default_project_id.create_instance(
            instance_id=CBT_INSTANCE,
            main_cluster_id=CBT_CLUSTER,
            main_cluster_zone=CBT_ZONE,
            replica_clusters=[{"id": CBT_REPLICA_CLUSTER_ID, "zone": CBT_REPLICA_CLUSTER_ZONE}],
            cluster_nodes=1,
            cluster_storage_type=enums.StorageType.SSD,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
            instance_type=enums.Instance.Type.DEVELOPMENT,
        )
        cluster.assert_has_calls(
            [
                mock.call(
                    cluster_id=CBT_CLUSTER, location_id=CBT_ZONE, default_storage_type=enums.StorageType.SSD
                ),
                mock.call(CBT_REPLICA_CLUSTER_ID, CBT_REPLICA_CLUSTER_ZONE, 1, enums.StorageType.SSD),
            ],
            any_order=True,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_create.assert_called_once_with(clusters=mock.ANY)
        assert res.instance_id == "instance"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("google.cloud.bigtable.instance.Instance.cluster")
    @mock.patch("google.cloud.bigtable.instance.Instance.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_instance_with_multiple_replica_clusters(
        self, get_client, instance_create, cluster, mock_project_id
    ):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_create.return_value = operation

        res = self.bigtable_hook_default_project_id.create_instance(
            instance_id=CBT_INSTANCE,
            main_cluster_id=CBT_CLUSTER,
            main_cluster_zone=CBT_ZONE,
            replica_clusters=CBT_REPLICATE_CLUSTERS,
            cluster_nodes=1,
            cluster_storage_type=enums.StorageType.SSD,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        cluster.assert_has_calls(
            [
                mock.call(
                    cluster_id=CBT_CLUSTER,
                    location_id=CBT_ZONE,
                    serve_nodes=1,
                    default_storage_type=enums.StorageType.SSD,
                ),
                mock.call("replica-1", "us-west1-a", 1, enums.StorageType.SSD),
                mock.call("replica-2", "us-central1-f", 1, enums.StorageType.SSD),
                mock.call("replica-3", "us-east1-d", 1, enums.StorageType.SSD),
            ],
            any_order=True,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_create.assert_called_once_with(clusters=mock.ANY)
        assert res.instance_id == "instance"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("google.cloud.bigtable.instance.Instance.update")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_update_instance(self, get_client, instance_update, mock_project_id):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_update.return_value = operation
        res = self.bigtable_hook_default_project_id.update_instance(
            instance_id=CBT_INSTANCE,
            instance_display_name=CBT_INSTANCE_DISPLAY_NAME,
            instance_type=CBT_INSTANCE_TYPE,
            instance_labels=CBT_INSTANCE_LABELS,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_update.assert_called_once_with()
        assert res.instance_id == "instance"

    @mock.patch("google.cloud.bigtable.instance.Instance.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_instance_overridden_project_id(self, get_client, instance_create):
        operation = mock.Mock()
        operation.result_return_value = Instance(instance_id=CBT_INSTANCE, client=get_client)
        instance_create.return_value = operation
        res = self.bigtable_hook_default_project_id.create_instance(
            project_id="new-project",
            instance_id=CBT_INSTANCE,
            main_cluster_id=CBT_CLUSTER,
            main_cluster_zone=CBT_ZONE,
        )
        get_client.assert_called_once_with(project_id="new-project")
        instance_create.assert_called_once_with(clusters=mock.ANY)
        assert res.instance_id == "instance"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.project_id",
        new_callable=PropertyMock,
        return_value=GCP_PROJECT_ID_HOOK_UNIT_TEST,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_table(self, get_client, mock_project_id):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        table_delete_method = instance_method.return_value.table.return_value.delete
        instance_exists_method.return_value = True
        self.bigtable_hook_default_project_id.delete_table(
            instance_id=CBT_INSTANCE,
            table_id=CBT_TABLE,
            project_id=GCP_PROJECT_ID_HOOK_UNIT_TEST,
        )
        get_client.assert_called_once_with(project_id="example-project")
        instance_exists_method.assert_called_once_with()
        table_delete_method.assert_called_once_with()

    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_table_overridden_project_id(self, get_client):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        table_delete_method = instance_method.return_value.table.return_value.delete
        instance_exists_method.return_value = True
        self.bigtable_hook_default_project_id.delete_table(
            project_id="new-project", instance_id=CBT_INSTANCE, table_id=CBT_TABLE
        )
        get_client.assert_called_once_with(project_id="new-project")
        instance_exists_method.assert_called_once_with()
        table_delete_method.assert_called_once_with()

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_credentials_and_project_id"
    )
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_delete_table_when_no_table_exists(self, mock_get_client, mock_get_creds_and_project):
        mock_get_creds_and_project.return_value = (mock.Mock(), "test-project")
        mock_client = mock.Mock()
        mock_get_client.return_value = mock_client
        instance = mock_client.instance.return_value
        instance.exists.return_value = True
        table = instance.table.return_value
        table.delete.side_effect = google.api_core.exceptions.NotFound("Table not found")
        self.bigtable_hook_default_project_id.delete_table(
            instance_id=CBT_INSTANCE,
            table_id=CBT_TABLE,
        )
        mock_get_client.assert_called_once_with(project_id="test-project")
        instance.exists.assert_called_once_with()
        instance.table.assert_called_once_with(table_id=CBT_TABLE)
        table.delete.assert_called_once_with()

    @mock.patch("google.cloud.bigtable.table.Table.create")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_create_table(self, get_client, create):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        client = mock.Mock(Client)
        instance = google.cloud.bigtable.instance.Instance(instance_id=CBT_INSTANCE, client=client)
        self.bigtable_hook_default_project_id.create_table(instance=instance, table_id=CBT_TABLE)
        get_client.assert_not_called()
        create.assert_called_once_with([], {})

    @mock.patch("google.cloud.bigtable.cluster.Cluster.update")
    @mock.patch("google.cloud.bigtable.cluster.Cluster.reload")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_update_cluster(self, get_client, reload, update):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        client = mock.Mock(Client)
        instance = google.cloud.bigtable.instance.Instance(instance_id=CBT_INSTANCE, client=client)
        self.bigtable_hook_default_project_id.update_cluster(
            instance=instance, cluster_id=CBT_CLUSTER, nodes=4
        )
        get_client.assert_not_called()
        reload.assert_called_once_with()
        update.assert_called_once_with()

    @mock.patch("google.cloud.bigtable.cluster.Cluster.update")
    @mock.patch("google.cloud.bigtable.cluster.Cluster.reload")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_update_cluster_does_not_exist(self, get_client, reload, update):
        instance_method = get_client.return_value.instance
        instance_method = instance_method.return_value
        instance_exists_method = instance_method.exists
        instance_exists_method.return_value = True
        reload.side_effect = google.api_core.exceptions.NotFound("Cluster not found")
        client = mock.Mock(Client)
        instance = google.cloud.bigtable.instance.Instance(instance_id=CBT_INSTANCE, client=client)
        with pytest.raises(
            AirflowException,
            match=f"Dependency: cluster '{CBT_CLUSTER}' does not exist for instance '{CBT_INSTANCE}'.",
        ):
            self.bigtable_hook_default_project_id.update_cluster(
                instance=instance,
                cluster_id=CBT_CLUSTER,
                nodes=4,
            )
        get_client.assert_not_called()
        reload.assert_called_once_with()
        update.assert_not_called()

    @mock.patch("google.cloud.bigtable.table.Table.list_column_families")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_list_column_families(self, get_client, list_column_families):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        client = mock.Mock(Client)
        get_client.return_value = client
        instance = google.cloud.bigtable.instance.Instance(instance_id=CBT_INSTANCE, client=client)
        self.bigtable_hook_default_project_id.get_column_families_for_table(
            instance=instance, table_id=CBT_TABLE
        )
        get_client.assert_not_called()
        list_column_families.assert_called_once_with()

    @mock.patch("google.cloud.bigtable.table.Table.get_cluster_states")
    @mock.patch("airflow.providers.google.cloud.hooks.bigtable.BigtableHook._get_client")
    def test_get_cluster_states(self, get_client, get_cluster_states):
        instance_method = get_client.return_value.instance
        instance_exists_method = instance_method.return_value.exists
        instance_exists_method.return_value = True
        client = mock.Mock(Client)
        instance = google.cloud.bigtable.instance.Instance(instance_id=CBT_INSTANCE, client=client)
        self.bigtable_hook_default_project_id.get_cluster_states_for_table(
            instance=instance, table_id=CBT_TABLE
        )
        get_client.assert_not_called()
        get_cluster_states.assert_called_once_with()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


"""
Example Airflow DAG for Jobs on Ray operations.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.exceptions import AirflowOptionalProviderFeatureException
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.ray import (
    RayDeleteJobOperator,
    RayGetJobInfoOperator,
    RayListJobsOperator,
    RayStopJobOperator,
    RaySubmitJobOperator,
)
from airflow.providers.google.cloud.operators.vertex_ai.ray import (
    CreateRayClusterOperator,
    DeleteRayClusterOperator,
    GetRayClusterOperator,
)

try:
    from google.cloud.aiplatform.vertex_ray.util import resources
except ImportError:
    raise AirflowOptionalProviderFeatureException(
        "The ray provider is optional and requires the `google-cloud-aiplatform` package to be installed. "
    )
try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
DAG_ID = "ray_job_operations"
LOCATION = "us-central1"
JOB_ID = f"{DAG_ID}_{ENV_ID}".replace("-", "_")
WORKER_NODE_RESOURCES = resources.Resources(
    node_count=1,
)

with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    render_template_as_native_obj=True,
    tags=["example", "job", "ray"],
) as dag:
    create_ray_cluster = CreateRayClusterOperator(
        task_id="create_ray_cluster",
        project_id=PROJECT_ID,
        location=LOCATION,
        worker_node_types=[WORKER_NODE_RESOURCES],
        python_version="3.10",
        ray_version="2.33",
    )

    get_ray_cluster = GetRayClusterOperator(
        task_id="get_ray_cluster",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=create_ray_cluster.output["cluster_id"],
    )

    # [START how_to_ray_submit_job]
    submit_ray_job = RaySubmitJobOperator(
        task_id="submit_ray_job",
        cluster_address="{{ task_instance.xcom_pull(task_ids='get_ray_cluster')['dashboard_address'] }}",
        entrypoint="echo hi && sleep 105 && echo hi2",
        runtime_env={
            "pip": [
                "ray==2.33.0",
            ],
        },
        get_job_logs=False,
        wait_for_job_done=False,
        submission_id=JOB_ID,
    )
    # [END how_to_ray_submit_job]

    # [START how_to_ray_get_job_info]
    info_ray_job = RayGetJobInfoOperator(
        task_id="info_ray_job",
        cluster_address="{{ task_instance.xcom_pull(task_ids='get_ray_cluster')['dashboard_address'] }}",
        job_id=JOB_ID,
    )
    # [END how_to_ray_get_job_info]

    # [START how_to_ray_list_jobs]
    list_ray_job = RayListJobsOperator(
        task_id="list_ray_job",
        cluster_address="{{ task_instance.xcom_pull(task_ids='get_ray_cluster')['dashboard_address'] }}",
    )
    # [END how_to_ray_list_jobs]

    # [START how_to_ray_stop_job]
    stop_ray_job = RayStopJobOperator(
        task_id="stop_ray_job",
        job_id=JOB_ID,
        cluster_address="{{ task_instance.xcom_pull(task_ids='get_ray_cluster')['dashboard_address'] }}",
    )
    # [END how_to_ray_stop_job]

    # [START how_to_ray_delete_job]
    delete_ray_job = RayDeleteJobOperator(
        task_id="delete_ray_job",
        cluster_address="{{ task_instance.xcom_pull(task_ids='get_ray_cluster')['dashboard_address'] }}",
        job_id=JOB_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END how_to_ray_delete_job]

    delete_ray_cluster = DeleteRayClusterOperator(
        task_id="delete_ray_cluster",
        project_id=PROJECT_ID,
        location=LOCATION,
        cluster_id=create_ray_cluster.output["cluster_id"],
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        create_ray_cluster
        >> get_ray_cluster
        >> submit_ray_job
        >> info_ray_job
        >> stop_ray_job
        >> list_ray_job
        >> delete_ray_job
        >> delete_ray_cluster
    )

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import AsyncMock

import pytest
from google.api_core.client_options import ClientOptions
from google.cloud.run_v2 import (
    CreateJobRequest,
    CreateServiceRequest,
    DeleteJobRequest,
    DeleteServiceRequest,
    GetJobRequest,
    GetServiceRequest,
    Job,
    JobsClient,
    ListJobsRequest,
    RunJobRequest,
    Service,
    ServicesAsyncClient,
    UpdateJobRequest,
)
from google.longrunning import operations_pb2

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.hooks.cloud_run import (
    CloudRunAsyncHook,
    CloudRunHook,
    CloudRunServiceAsyncHook,
    CloudRunServiceHook,
)

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_default_project_id

PROJECT_ID = "projectid"
REGION = "region1"
JOB_NAME = "job1"
SERVICE_NAME = "service1"
OPERATION_NAME = "operationname"
USE_REGIONAL_ENDPOINT = True
BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"


@pytest.mark.db_test
class TestCloudRunHook:
    def dummy_get_credentials(self):
        pass

    @pytest.fixture
    def cloud_run_hook(self):
        cloud_run_hook = CloudRunHook()
        cloud_run_hook.get_credentials = self.dummy_get_credentials
        return cloud_run_hook

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_get_conn_regional_endpoint(self, mock_jobs_client_cls):
        hook = CloudRunHook(gcp_conn_id="google_cloud_default")
        hook.get_credentials = mock.MagicMock(return_value=mock.Mock())

        location = "us-central1"
        hook.get_conn(location=location, use_regional_endpoint=USE_REGIONAL_ENDPOINT)
        assert mock_jobs_client_cls.call_count == 1

        _, kwargs = mock_jobs_client_cls.call_args
        client_options = kwargs.get("client_options")
        assert isinstance(client_options, ClientOptions)
        assert client_options.api_endpoint == f"{location}-run.googleapis.com:443"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_get_job(self, mock_batch_service_client, cloud_run_hook):
        get_job_request = GetJobRequest(name=f"projects/{PROJECT_ID}/locations/{REGION}/jobs/{JOB_NAME}")

        cloud_run_hook.get_job(
            job_name=JOB_NAME,
            region=REGION,
            project_id=PROJECT_ID,
            use_regional_endpoint=USE_REGIONAL_ENDPOINT,
        )
        cloud_run_hook._client.get_job.assert_called_once_with(get_job_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_update_job(self, mock_batch_service_client, cloud_run_hook):
        job = Job()
        job.name = f"projects/{PROJECT_ID}/locations/{REGION}/jobs/{JOB_NAME}"

        update_request = UpdateJobRequest()
        update_request.job = job

        cloud_run_hook.update_job(
            job=Job.to_dict(job),
            job_name=JOB_NAME,
            region=REGION,
            project_id=PROJECT_ID,
            use_regional_endpoint=USE_REGIONAL_ENDPOINT,
        )

        cloud_run_hook._client.update_job.assert_called_once_with(update_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_create_job(self, mock_batch_service_client, cloud_run_hook):
        job = Job()

        create_request = CreateJobRequest()
        create_request.job = job
        create_request.job_id = JOB_NAME
        create_request.parent = f"projects/{PROJECT_ID}/locations/{REGION}"

        cloud_run_hook.create_job(
            job=Job.to_dict(job),
            job_name=JOB_NAME,
            region=REGION,
            project_id=PROJECT_ID,
            use_regional_endpoint=USE_REGIONAL_ENDPOINT,
        )

        cloud_run_hook._client.create_job.assert_called_once_with(create_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_execute_job(self, mock_batch_service_client, cloud_run_hook):
        overrides = {
            "container_overrides": [{"args": ["python", "main.py"]}],
            "task_count": 1,
            "timeout": "60s",
        }
        run_job_request = RunJobRequest(
            name=f"projects/{PROJECT_ID}/locations/{REGION}/jobs/{JOB_NAME}", overrides=overrides
        )

        cloud_run_hook.execute_job(
            job_name=JOB_NAME,
            region=REGION,
            project_id=PROJECT_ID,
            overrides=overrides,
            use_regional_endpoint=USE_REGIONAL_ENDPOINT,
        )
        cloud_run_hook._client.run_job.assert_called_once_with(request=run_job_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_list_jobs(self, mock_batch_service_client, cloud_run_hook):
        number_of_jobs = 3
        region = "us-central1"
        project_id = "test_project_id"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_run_hook.list_jobs(
            region=region, project_id=project_id, use_regional_endpoint=USE_REGIONAL_ENDPOINT
        )

        for i in range(number_of_jobs):
            assert jobs_list[i].name == f"name{i}"

        expected_list_jobs_request: ListJobsRequest = ListJobsRequest(
            parent=f"projects/{project_id}/locations/{region}"
        )
        mock_batch_service_client.return_value.list_jobs.assert_called_once_with(
            request=expected_list_jobs_request
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_list_jobs_show_deleted(self, mock_batch_service_client, cloud_run_hook):
        number_of_jobs = 3
        region = "us-central1"
        project_id = "test_project_id"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_run_hook.list_jobs(
            region=region,
            project_id=project_id,
            show_deleted=True,
            use_regional_endpoint=USE_REGIONAL_ENDPOINT,
        )

        for i in range(number_of_jobs):
            assert jobs_list[i].name == f"name{i}"

        expected_list_jobs_request: ListJobsRequest = ListJobsRequest(
            parent=f"projects/{project_id}/locations/{region}", show_deleted=True
        )
        mock_batch_service_client.return_value.list_jobs.assert_called_once_with(
            request=expected_list_jobs_request
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_list_jobs_with_limit(self, mock_batch_service_client, cloud_run_hook):
        number_of_jobs = 3
        limit = 2
        region = "us-central1"
        project_id = "test_project_id"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_run_hook.list_jobs(
            region=region, project_id=project_id, limit=limit, use_regional_endpoint=USE_REGIONAL_ENDPOINT
        )

        assert len(jobs_list) == limit
        for i in range(limit):
            assert jobs_list[i].name == f"name{i}"

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_list_jobs_with_limit_zero(self, mock_batch_service_client, cloud_run_hook):
        number_of_jobs = 3
        limit = 0
        region = "us-central1"
        project_id = "test_project_id"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_run_hook.list_jobs(
            region=region, project_id=project_id, limit=limit, use_regional_endpoint=USE_REGIONAL_ENDPOINT
        )

        assert len(jobs_list) == 0

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_list_jobs_with_limit_greater_then_range(self, mock_batch_service_client, cloud_run_hook):
        number_of_jobs = 3
        limit = 5
        region = "us-central1"
        project_id = "test_project_id"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_run_hook.list_jobs(
            region=region, project_id=project_id, limit=limit, use_regional_endpoint=USE_REGIONAL_ENDPOINT
        )

        assert len(jobs_list) == number_of_jobs
        for i in range(number_of_jobs):
            assert jobs_list[i].name == f"name{i}"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_list_jobs_with_limit_less_than_zero(self, mock_batch_service_client, cloud_run_hook):
        number_of_jobs = 3
        limit = -1
        region = "us-central1"
        project_id = "test_project_id"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        with pytest.raises(expected_exception=AirflowException):
            cloud_run_hook.list_jobs(
                region=region, project_id=project_id, limit=limit, use_regional_endpoint=USE_REGIONAL_ENDPOINT
            )

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_delete_job(self, mock_batch_service_client, cloud_run_hook):
        delete_request = DeleteJobRequest(name=f"projects/{PROJECT_ID}/locations/{REGION}/jobs/{JOB_NAME}")

        cloud_run_hook.delete_job(
            job_name=JOB_NAME,
            region=REGION,
            project_id=PROJECT_ID,
            use_regional_endpoint=USE_REGIONAL_ENDPOINT,
        )
        cloud_run_hook._client.delete_job.assert_called_once_with(delete_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_get_conn_with_transport(self, mock_jobs_client):
        """Test that transport parameter is passed to JobsClient."""
        hook = CloudRunHook(transport="rest")
        hook.get_credentials = self.dummy_get_credentials
        hook.get_conn(location=REGION, use_regional_endpoint=USE_REGIONAL_ENDPOINT)

        mock_jobs_client.assert_called_once()
        call_kwargs = mock_jobs_client.call_args[1]
        assert call_kwargs["transport"] == "rest"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_get_conn_omits_transport_when_none(self, mock_jobs_client):
        """Test that transport is not passed to JobsClient when None."""
        hook = CloudRunHook(transport=None)
        hook.get_credentials = self.dummy_get_credentials
        hook.get_conn(location=REGION, use_regional_endpoint=USE_REGIONAL_ENDPOINT)

        mock_jobs_client.assert_called_once()
        call_kwargs = mock_jobs_client.call_args[1]
        assert "transport" not in call_kwargs

    def _mock_pager(self, number_of_jobs):
        mock_pager = []
        for i in range(number_of_jobs):
            mock_pager.append(Job(name=f"name{i}"))

        return mock_pager


class TestCloudRunAsyncHook:
    @pytest.mark.asyncio
    async def test_get_operation(self):
        region = "us-central1"
        hook = CloudRunAsyncHook()
        hook.get_conn = mock.AsyncMock()
        await hook.get_operation(
            operation_name=OPERATION_NAME, location=region, use_regional_endpoint=USE_REGIONAL_ENDPOINT
        )
        hook.get_conn.return_value.get_operation.assert_called_once_with(
            operations_pb2.GetOperationRequest(name=OPERATION_NAME), timeout=120
        )

    @pytest.mark.asyncio
    @pytest.mark.parametrize("transport", [None, "grpc"])
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsAsyncClient")
    async def test_get_conn_uses_async_client_by_default(self, mock_async_client, transport):
        """Test that get_conn uses JobsAsyncClient (grpc_asyncio) when transport is None or grpc."""
        hook = CloudRunAsyncHook(transport=transport)
        mock_sync_hook = mock.MagicMock(spec=CloudRunHook)
        mock_sync_hook.get_credentials.return_value = "credentials"
        hook.get_sync_hook = mock.AsyncMock(return_value=mock_sync_hook)

        await hook.get_conn(location=REGION, use_regional_endpoint=USE_REGIONAL_ENDPOINT)

        mock_async_client.assert_called_once()
        call_kwargs = mock_async_client.call_args[1]
        assert "transport" not in call_kwargs

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    async def test_get_conn_uses_sync_client_for_rest(self, mock_sync_client):
        """Test that get_conn uses sync JobsClient with REST transport."""
        hook = CloudRunAsyncHook(transport="rest")
        mock_sync_hook = mock.MagicMock(spec=CloudRunHook)
        mock_sync_hook.get_credentials.return_value = "credentials"
        hook.get_sync_hook = mock.AsyncMock(return_value=mock_sync_hook)

        await hook.get_conn(location=REGION, use_regional_endpoint=USE_REGIONAL_ENDPOINT)

        mock_sync_client.assert_called_once()
        call_kwargs = mock_sync_client.call_args[1]
        assert call_kwargs["transport"] == "rest"

    @pytest.mark.asyncio
    @mock.patch("asyncio.to_thread")
    async def test_get_operation_rest_uses_to_thread(self, mock_to_thread):
        """Test that get_operation uses asyncio.to_thread for REST transport."""
        expected_operation = operations_pb2.Operation(name=OPERATION_NAME)
        mock_to_thread.return_value = expected_operation

        hook = CloudRunAsyncHook(transport="rest")
        mock_conn = mock.MagicMock(spec=JobsClient)  # sync client
        hook.get_conn = mock.AsyncMock(return_value=mock_conn)

        result = await hook.get_operation(
            operation_name=OPERATION_NAME, location=REGION, use_regional_endpoint=USE_REGIONAL_ENDPOINT
        )

        mock_to_thread.assert_called_once_with(
            mock_conn.get_operation,
            operations_pb2.GetOperationRequest(name=OPERATION_NAME),
            timeout=120,
        )
        assert result == expected_operation


@pytest.mark.db_test
class TestCloudRunServiceHook:
    def dummy_get_credentials(self):
        pass

    @pytest.fixture
    def cloud_run_service_hook(self):
        region = "us-central1"
        cloud_run_service_hook = CloudRunServiceHook()
        cloud_run_service_hook.get_credentials = self.dummy_get_credentials
        cloud_run_service_hook.client_options = ClientOptions(api_endpoint=f"{region}-run.googleapis.com:443")
        return cloud_run_service_hook

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseAsyncHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.ServicesClient")
    def test_get_service(self, mock_batch_service_client, cloud_run_service_hook):
        get_service_request = GetServiceRequest(
            name=f"projects/{PROJECT_ID}/locations/{REGION}/services/{SERVICE_NAME}"
        )

        cloud_run_service_hook.get_service(
            service_name=SERVICE_NAME,
            region=REGION,
            project_id=PROJECT_ID,
            use_regional_endpoint=USE_REGIONAL_ENDPOINT,
        )
        cloud_run_service_hook._client.get_service.assert_called_once_with(get_service_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseAsyncHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.ServicesClient")
    def test_create_service(self, mock_batch_service_client, cloud_run_service_hook):
        service = Service()

        create_request = CreateServiceRequest()
        create_request.service = service
        create_request.service_id = SERVICE_NAME
        create_request.parent = f"projects/{PROJECT_ID}/locations/{REGION}"

        cloud_run_service_hook.create_service(
            service=service,
            service_name=SERVICE_NAME,
            region=REGION,
            project_id=PROJECT_ID,
            use_regional_endpoint=USE_REGIONAL_ENDPOINT,
        )
        cloud_run_service_hook._client.create_service.assert_called_once_with(create_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseAsyncHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.ServicesClient")
    def test_delete_service(self, mock_batch_service_client, cloud_run_service_hook):
        delete_request = DeleteServiceRequest(
            name=f"projects/{PROJECT_ID}/locations/{REGION}/services/{SERVICE_NAME}"
        )

        cloud_run_service_hook.delete_service(
            service_name=SERVICE_NAME,
            region=REGION,
            project_id=PROJECT_ID,
            use_regional_endpoint=USE_REGIONAL_ENDPOINT,
        )
        cloud_run_service_hook._client.delete_service.assert_called_once_with(delete_request)


class TestCloudRunServiceAsyncHook:
    def dummy_get_credentials(self):
        pass

    def mock_service(self):
        return mock.AsyncMock()

    @pytest.fixture
    def cloud_run_service_hook(self):
        region = "us-central1"
        cloud_run_service_hook = CloudRunServiceAsyncHook()
        cloud_run_service_hook.get_credentials = self.dummy_get_credentials
        cloud_run_service_hook.client_options = ClientOptions(api_endpoint=f"{region}-run.googleapis.com:443")
        return cloud_run_service_hook

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseAsyncHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.CloudRunServiceAsyncHook.get_conn")
    async def test_create_service(self, mock_client, cloud_run_service_hook):
        mock_env_client = AsyncMock(ServicesAsyncClient)
        mock_client.return_value = mock_env_client

        await cloud_run_service_hook.create_service(
            service_name=SERVICE_NAME,
            service=Service(),
            region=REGION,
            project_id=PROJECT_ID,
            use_regional_endpoint=USE_REGIONAL_ENDPOINT,
        )

        expected_request = CreateServiceRequest(
            service=Service(),
            service_id=SERVICE_NAME,
            parent=f"projects/{PROJECT_ID}/locations/{REGION}",
        )

        mock_client.return_value.create_service.assert_called_once_with(expected_request)

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseAsyncHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.CloudRunServiceAsyncHook.get_conn")
    async def test_delete_service(self, mock_client, cloud_run_service_hook):
        mock_env_client = AsyncMock(ServicesAsyncClient)
        mock_client.return_value = mock_env_client

        await cloud_run_service_hook.delete_service(
            service_name=SERVICE_NAME,
            region=REGION,
            project_id=PROJECT_ID,
            use_regional_endpoint=USE_REGIONAL_ENDPOINT,
        )

        expected_request = DeleteServiceRequest(
            name=f"projects/{PROJECT_ID}/locations/{REGION}/services/{SERVICE_NAME}",
        )

        mock_client.return_value.delete_service.assert_called_once_with(expected_request)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Vertex AI Generative AI operators."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.providers.google.cloud.hooks.vertex_ai.generative_model import (
    GenerativeModelHook,
)
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class RunEvaluationOperator(GoogleCloudBaseOperator):
    """
    Use the Rapid Evaluation API to evaluate a model.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param pretrained_model: Required. A pre-trained model optimized for performing natural
        language tasks such as classification, summarization, extraction, content
        creation, and ideation.
    :param eval_dataset: Required. A fixed dataset for evaluating a model against. Adheres to Rapid Evaluation API.
    :param metrics: Required. A list of evaluation metrics to be used in the experiment. Adheres to Rapid Evaluation API.
    :param experiment_name: Required. The name of the evaluation experiment.
    :param experiment_run_name: Required. The specific run name or ID for this experiment.
    :param prompt_template: Required. The template used to format the model's prompts during evaluation. Adheres to Rapid Evaluation API.
    :param generation_config: Optional. A dictionary containing generation parameters for the model.
    :param safety_settings: Optional. A dictionary specifying harm category thresholds for blocking model outputs.
    :param system_instruction: Optional. An instruction given to the model to guide its behavior.
    :param tools: Optional. A list of tools available to the model during evaluation, such as a data store.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "pretrained_model",
        "eval_dataset",
        "prompt_template",
        "experiment_name",
        "experiment_run_name",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        pretrained_model: str,
        eval_dataset: dict,
        metrics: list,
        experiment_name: str,
        experiment_run_name: str,
        prompt_template: str,
        generation_config: dict | None = None,
        safety_settings: dict | None = None,
        system_instruction: str | None = None,
        tools: list | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.pretrained_model = pretrained_model
        self.eval_dataset = eval_dataset
        self.metrics = metrics
        self.experiment_name = experiment_name
        self.experiment_run_name = experiment_run_name
        self.prompt_template = prompt_template
        self.generation_config = generation_config
        self.safety_settings = safety_settings
        self.system_instruction = system_instruction
        self.tools = tools
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = self.hook.run_evaluation(
            project_id=self.project_id,
            location=self.location,
            pretrained_model=self.pretrained_model,
            eval_dataset=self.eval_dataset,
            metrics=self.metrics,
            experiment_name=self.experiment_name,
            experiment_run_name=self.experiment_run_name,
            prompt_template=self.prompt_template,
            generation_config=self.generation_config,
            safety_settings=self.safety_settings,
            system_instruction=self.system_instruction,
            tools=self.tools,
        )

        return response.summary_metrics



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.google.cloud.triggers.pubsub import PubsubPullTrigger

pytest.importorskip("airflow.providers.common.messaging.providers.base_provider")


def test_message_pubsub_queue_create():
    from airflow.providers.common.messaging.providers.base_provider import BaseMessageQueueProvider
    from airflow.providers.google.event_scheduling.events.pubsub import (
        PubSubMessageQueueEventTriggerContainer,
    )

    provider = PubSubMessageQueueEventTriggerContainer()
    assert isinstance(provider, BaseMessageQueueProvider)


def test_message_pubsub_queue_trigger_class():
    from airflow.providers.google.event_scheduling.events.pubsub import (
        PubSubMessageQueueEventTriggerContainer,
    )

    provider = PubSubMessageQueueEventTriggerContainer()
    assert provider.trigger_class() == PubsubPullTrigger


def test_scheme_matches():
    from airflow.providers.google.event_scheduling.events.pubsub import (
        PubSubMessageQueueEventTriggerContainer,
    )

    provider = PubSubMessageQueueEventTriggerContainer()
    assert provider.scheme_matches("google+pubsub")
